package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.physics.Poly
import ru.dbotthepony.kvector.vector.Vector2d

abstract class Entity(val world: World<*, *>) {
	/**
	 * The chunk this entity currently in, it is automatically updated on each change of [position]
	 */
	var chunk: Chunk<*, *>? = null
		set(value) {
			if (!isSpawned) {
				throw IllegalStateException("Trying to set chunk this entity belong to before spawning in world")
			} else if (isRemoved) {
				throw IllegalStateException("This entity was removed")
			} else if (value == field) {
				return
			}

			val chunkPos = world.chunkFromCell(position)

			if (value != null && chunkPos != value.pos) {
				throw IllegalStateException("Set proper position before setting chunk this Entity belongs to (expected chunk $chunkPos, got chunk ${value.pos})")
			}

			val oldChunk = field
			field = value

			if (oldChunk == null && value != null) {
				world.orphanedEntities.remove(this)
				value.addEntity(this)
			} else if (oldChunk != null && value == null) {
				world.orphanedEntities.add(this)
				oldChunk.removeEntity(this)
			} else if (oldChunk != null && value != null) {
				value.transferEntity(this, oldChunk)
			}
		}


	/**
	 * Current entity position. If entity is logical, this can only be changed
	 * by external means, otherwise it always stands where it is.
	 *
	 * If entity is physical, then movement controller will update this on each physics step
	 * to match position of physical body.
	 *
	 * Setting this value will update [chunk] immediately, if [isSpawned] is true and [isRemoved] is false.
	 */
	var position = Vector2d()
		set(value) {
			if (field == value)
				return

			val old = field
			field = Vector2d(world.x.cell(value.x), world.y.cell(value.y))

			if (isSpawned && !isRemoved) {
				val oldChunkPos = world.chunkFromCell(old)
				val newChunkPos = world.chunkFromCell(field)

				if (oldChunkPos != newChunkPos) {
					chunk = world.chunkMap[newChunkPos]
				}
			}
		}


	/**
	 * This entity's angle in radians.
	 *
	 * Logical entities never rotate.
	 *
	 * Alive entities usually don't rotate.
	 *
	 * If entity is physical, this value is updated by movement controller on
	 * each physics step to match angle of physical body.
	 */
	var angle: Double = 0.0
		set(value) {
			if (field == value)
				return

			field = value
		}

	var velocity = Vector2d.ZERO
	val hitboxes = ArrayList<Poly>()

	/**
	 * Whenever is this entity spawned in world ([spawn] called).
	 * Doesn't mean entity still exists in world, check it with [isRemoved]
	 */
	var isSpawned = false
		private set

	/**
	 * Whenever is this entity was removed from world ([remove] called).
	 */
	var isRemoved = false
		private set

	open fun spawn() {
		if (isSpawned)
			throw IllegalStateException("Already spawned")

		isSpawned = true
		world.entities.add(this)
		chunk = world.chunkMap[world.chunkFromCell(position)]

		if (chunk == null) {
			world.orphanedEntities.add(this)
		}
	}

	open fun remove() {
		if (isRemoved)
			throw IllegalStateException("Already removed")

		isRemoved = true

		if (isSpawned) {
			world.entities.remove(this)
			chunk?.removeEntity(this)
		}
	}

	/**
	 * Заставляет сущность "думать".
	 */
	fun think() {
		if (!isSpawned) {
			throw IllegalStateException("Tried to think before spawning in world")
		}

		move()
		thinkInner()
	}

	protected abstract fun thinkInner()

	protected open fun move() {
		position += velocity

		val polies = ArrayList<Poly>()

		for (x in -1 .. 1) {
			for (y in -1 .. 1) {
				world.chunkMap.getCell(position.x.toInt() + x, position.y.toInt() + y)?.let {
					polies.addAll(it.polies)
				}
			}
		}

		for (i in 0 until 10) {
			val intersects = ArrayList<Poly.Penetration>()

			this.hitboxes.forEach { hitbox0 ->
				val hitbox = hitbox0 + position
				polies.forEach { poly -> hitbox.intersect(poly)?.let { intersects.add(it) } }
			}

			if (intersects.isEmpty())
				break
			else {
				val max = intersects.max()
				// resolve collision
				position += max.vector
				// collision response
				velocity -= max.axis * velocity.dot(max.axis * 1.1)
			}
		}
	}

	open fun onTouchSurface(velocity: Vector2d, normal: Vector2d) {

	}

	open fun render(client: StarboundClient = StarboundClient.current()) {
		hitboxes.forEach { (it + position).render(client) }
	}

	open var maxHealth = 0.0
	open var health = 0.0

	open fun hurt(amount: Double): Boolean {
		return false
	}
}
